package descartes.kieker4dql.dqlconnector;

import java.io.File;
import java.io.IOException;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import org.apache.log4j.Logger;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.emf.ecore.util.EcoreUtil;

import descartes.kieker4dql.dqlconnector.structures.Resource;
import descartes.kieker4dql.dqlconnector.structures.Trace;
import descartes.kieker4dql.dqlconnector.structures.identifier.ResourceIdentifier;
import descartes.kieker4dql.dqlconnector.structures.records.ResourceRecord;
import edu.kit.ipd.descartes.mm.usageprofile.UsageProfilePackage;
import edu.kit.ipd.descartes.mm.usageprofile.UsageScenario;
import edu.kit.ipd.descartes.mm.usageprofile.WorkloadType;
import edu.kit.ipd.descartes.mm.usageprofile.impl.ClosedWorkloadTypeImpl;
import edu.kit.ipd.descartes.mm.usageprofile.impl.OpenWorkloadTypeImpl;
import edu.kit.ipd.descartes.mm.usageprofile.impl.UsageScenarioImpl;
import edu.kit.ipd.descartes.solving.DMLModels;
import edu.kit.ipd.descartes.solving.DMLSolver;
import edu.kit.ipd.descartes.solving.ModelValidator;
import edu.kit.ipd.descartes.solving.SimulationConfig;
import edu.kit.ipd.descartes.solving.probes.AvgThroughputProbe;
import edu.kit.ipd.descartes.solving.probes.ProbeRepository;
import edu.kit.ipd.descartes.solving.probes.ProbesFactory;
import tools.descartes.dql.connector.AbstractOSGiQueryConnector;
import tools.descartes.dql.connector.PerformanceMetricsQueryConnector;
import tools.descartes.dql.connector.QueryProgressListener;
import tools.descartes.dql.core.engine.util.DQLLogger;
import tools.descartes.dql.models.mapping.domain.ConfidenceInterval;
import tools.descartes.dql.models.mapping.domain.DecimalResult;
import tools.descartes.dql.models.mapping.domain.DoF;
import tools.descartes.dql.models.mapping.domain.DomainFactory;
import tools.descartes.dql.models.mapping.domain.Entity;
import tools.descartes.dql.models.mapping.domain.Probe;
import tools.descartes.dql.models.mapping.domain.impl.ConfidenceIntervalImpl;
import tools.descartes.dql.models.mapping.mapping.EntityMapping;

public class PerformanceMetricsQueryConnectorImpl extends
		AbstractOSGiQueryConnector implements PerformanceMetricsQueryConnector {

	private final Logger log = DQLLogger.getLogger(this.getClass().getName());
	ArrayList<String> constraintNames = new ArrayList<String>();
	ArrayList<String> doFConstraintNames = new ArrayList<String>();
	ArrayList<String> explorationStrategyNames = new ArrayList<String>();

	private static final String[] SUPPORTED_CONSTRAINTS = {"ACCURATE", "BALANCED", "FAST"};
	private int constraintIndex = 0;

	private String constraint = null;
	private String doFConstraint = null;
	private String exploratationStrategy = null;
	private Properties explorationStrategyProperties;

	private final DomainFactory domainFactory = DomainFactory.eINSTANCE;

	public PerformanceMetricsQueryConnectorImpl() {
		super();

		loadConstraintNames();
		loadDoFConstraintNames();
		loadExplorationStrategyNames();
	}

	private void loadConstraintNames() {
		constraintNames.add("DEFAULT");
		constraintNames.add("FAST");
		constraintNames.add("ACCURATE");
		constraintNames.add("TIME_1s");
		constraintNames.add("TIME_5s");
		constraintNames.add("ACCURACY_95");
		constraintNames.add("ACCURACY_99");
	}

	private void loadDoFConstraintNames() {
		doFConstraintNames.add("DEFAULT");
		doFConstraintNames.add("FAST");
		doFConstraintNames.add("ACCURATE");
		doFConstraintNames.add("TIME_1s");
		doFConstraintNames.add("TIME_5s");
		doFConstraintNames.add("ACCURACY_95");
		doFConstraintNames.add("ACCURACY_99");
	}

	private void loadExplorationStrategyNames() {
		explorationStrategyNames.add("FULL");
		explorationStrategyNames.add("RANDOM");
	}

	// @Override
	// public String getConstraint() {
	// return SUPPORTED_CONSTRAINTS[this.constraintIndex];
	// }

	@Override
	public String getConstraint() {
		return constraint;
	}

	@Override
	public String getDoFConstraint() {
		return doFConstraint;
	}

	@Override
	public String getDoFExplorationStrategy() {
		return exploratationStrategy;
	}

	@Override
	public Properties getDoFExplorationStrategyProperties() {
		return null;
	}

	@Override
	public List<String> getSupportedConstraintNames() {
		return Arrays.asList(SUPPORTED_CONSTRAINTS);
	}

	@Override
	public List<String> getSupportedDoFConstraintNames() {
		return null;
	}

	@Override
	public List<String> getSupportedExplorationStrategyNames() {
		return null;
	}

	@Override
	public List<String> getSupportedTimeUnits() {
		// Unsupported
		return null;
	}
	

	/* Modifies response to include the results of Kieker */
	private void processKiekerSolverResults(EntityMapping response, Trace kiekerResults) {

		// Map containing the resulting records sorted by resources
		final HashMap<String, List<ResourceRecord>> kiekerResultMap = new HashMap<String, List<ResourceRecord>>(
				kiekerResults.getResourceIdentifierToResourceMap().size());

		// List of Records for one resource, value for kiekerResultMap
		List<ResourceRecord> records = new ArrayList<>();
		records = kiekerResults.getResources();

		// kieker resource id to resource mapping, help for identifying the
		// records
		Map<ResourceIdentifier, Resource> resourceMap = kiekerResults.getResourceIdentifierToResourceMap();

		// for every retrieved record get the original dql identifier, if no
		// match found but still in map, throw away)
		// TODO: should be avoided by resource filter
		for (ResourceRecord kiekerProbe : records) {
			if (resourceMap.get(kiekerProbe.getIdentifier()) != null) {
				String identifier = resourceMap.get(kiekerProbe.getIdentifier()).getName();
				List<ResourceRecord> kiekerProbes = kiekerResultMap.get(identifier);
				if (null == kiekerProbes) {
					kiekerProbes = new ArrayList<ResourceRecord>();
					kiekerProbes.add(kiekerProbe);
					kiekerResultMap.put(identifier, kiekerProbes);

				} else {
					kiekerProbes.add(kiekerProbe);
					kiekerResultMap.put(identifier, kiekerProbes);

				}
			}

		}
		
		final ArrayList<Entity> entities = new ArrayList<Entity>();
		entities.addAll(response.getResources());
		entities.addAll(response.getServices());
		ArrayList<Probe> helper = null;
		
		// browse through each entity and its probes 
		//   and try to replace the probe with a result
		for (Entity ent : entities) {
			helper = new ArrayList<Probe>(ent.getProbes().size());
			for (Probe dqlProbe : ent.getProbes()) {
				System.out.println("ORIGIN ID: " + ent.getIdentifier().toString());
				List<ResourceRecord> kiekerProbes = kiekerResultMap.get( ent.getIdentifier().toString());
				String value = null;
				if (kiekerProbes != null) {
					if (DMLHelper.METRIC_RESPONSE_TIME_AVG.equalsIgnoreCase(dqlProbe.getMetricName())) {
						for (ResourceRecord kiekerProbe : kiekerProbes) {
							value = kiekerProbe.toString();
							break;
						}
					} else if (DMLHelper.METRIC_UTILIZATION.equalsIgnoreCase(dqlProbe.getMetricName())) {
						for (ResourceRecord kiekerProbe : kiekerProbes) {
							if (true) {
								value = String.valueOf(kiekerProbe.getContent().get(2));
								break;
							}
						}
					} else if (DMLHelper.METRIC_THROUGHPUT.equalsIgnoreCase(dqlProbe.getMetricName())) {
						for (ResourceRecord dmlProbe : kiekerProbes) {
							if (dmlProbe instanceof AvgThroughputProbe) {
								value = ((AvgThroughputProbe) dmlProbe).getValue();
								break;
							}
						}
					} else {
						// do nothing
					}
				}
				
				if (null != value) {
					DecimalResult res = domainFactory.createDecimalResult();
					res.setMetricName(dqlProbe.getMetricName());
					ConfidenceInterval confidenceInterval = new ConfidenceIntervalImpl();
					confidenceInterval.setAccuracy(new BigDecimal(1));
					res.setConfidenceInterval(confidenceInterval);
					res.setValid(true);
					try {
						BigDecimal val = new BigDecimal(value);
						res.setValue(val);
					} catch (NumberFormatException exc) {
						res.setValid(false);
					}	
					helper.add(res);
				} else {
					helper.add(dqlProbe);
				}

			}
			ent.getProbes().clear();
			ent.getProbes().addAll(helper);
		}
	}

	@Override
	public List<EntityMapping> requestDoFMetrics(final EntityMapping request) {
		final ArrayList<EntityMapping> results = new ArrayList<EntityMapping>();
		final String modelLocation = request.getModelLocation();
		final DMLModelFiles dmlModelFiles = DMLHelper.createDMLModelFiles(modelLocation);

		// Load and validate Model
		DMLModels models = DMLHelper.loadModels(dmlModelFiles);
		ModelValidator validator = new ModelValidator(models);
		if (!validator.isSetupValid()) {
			DMLHelper.openValidationErrorWindow(validator.getErrors());
			return null;
		}

		for (final DoF doF : request.getDoF()) {
			Properties x = doF.getProperties();
			String arrayString = (String) x.get("value");
			arrayString = arrayString.replace("]", "");
			arrayString = arrayString.replace("[", "");
			arrayString = arrayString.replace(" ", "");
			String[] arr = arrayString.split(",");

			EObject dofObject = getServiceWithIdentifier(models, doF.getIdentifier());
			if (dofObject.getClass().equals(UsageScenarioImpl.class)) {
				WorkloadType workloadType = ((UsageScenario) dofObject).getWorkloadType();
				// "self.oclIsKindOf(ProcessingResourceSpecification)"
				if (workloadType.getClass().equals(OpenWorkloadTypeImpl.class)) {
				} else if (workloadType.getClass().equals(ClosedWorkloadTypeImpl.class)) {
					workloadType.eGet(UsageProfilePackage.Literals.CLOSED_WORKLOAD_TYPE__POPULATION);
					for (int i = 0; i < arr.length; i++) {
						workloadType.eSet(UsageProfilePackage.Literals.CLOSED_WORKLOAD_TYPE__POPULATION,
								new BigInteger(arr[i].trim()));
						System.out.println(arr[i].trim());
						System.out.println(
								workloadType.eGet(UsageProfilePackage.Literals.CLOSED_WORKLOAD_TYPE__POPULATION));
						// TODO start analysis
						// processDMLSolverResults(response, dmlResults);
						// response.s
					}
				}
			}

			// queryResults.add(workloadType);
			// Setting abcd;
			// EStructuralFeature
			// Object xasdasd =
			// workloadType.eGet(UsageProfilePackage.Literals.CLOSED_WORKLOAD_TYPE__POPULATION);
			// TODO see AdaptableEntity;

			// queryResults.add(ClosedWorkloadType);
			// workloadType).getPopulation());
			// RandomVariable thinktime = ((ClosedWorkloadType)
			// workloadType).getThinkTime();
			// queryResults.add(thinktime);
			// ProbabilityFunction probabilityFunction =
			// ((ClosedWorkloadType) workloadType).getThinkTime()
			// .getProbFunction();
			// queryResults.add(probabilityFunction);
			// results.add(requestMetrics(request));
		}

		return results;
	}

	@Override
	public EntityMapping requestMetrics(final EntityMapping request) {
		System.out.println("Starting Kieker Controller now" + request.toString());

		final EntityMapping response = EcoreUtil.copy(request);
		final String modelLocation = request.getModelLocation();

		// Start KiekerController with custom config (currently only location of
		// traceFiles)
		KiekerController c = new KiekerController(modelLocation + "//controller.cfg");

		// Fill the traces according to the applied filters
		c.fillTrace();

		// Get the result trace
		Trace result = c.getTrace();

		// Process the trace back to a DQL result
		processKiekerSolverResults(response, result);
		return response;
	}
	
	private EObject getResourceWithIdentifier(DMLModels models, String identifier) {
		// resources are located in the resource landscape
		final List<EObject> queryResults = DMLHelper.executeOCLQueryWithoutContext(
				new EObject[] { models.getResourceLandscapeModel() },
				"self.oclIsKindOf(identifier::Identifier) and self.oclAsType(identifier::Identifier).id='" + identifier + "'");
		assert queryResults != null && queryResults.size() < 2; // at most one matching resource can be found
		return queryResults.isEmpty() ? null : queryResults.get(0);
	}

	private EObject getServiceWithIdentifier(DMLModels models, String identifier) {
		// services can be located either in the usage profile or in the repository
		List<EObject> queryResults = DMLHelper.executeOCLQueryWithoutContext(
				new EObject[] { models.getUsageProfileModel() },
				"self.oclIsKindOf(identifier::Identifier) and self.oclAsType(identifier::Identifier).id='" + identifier + "'");
		if (null == queryResults || queryResults.isEmpty()) {
			queryResults = DMLHelper.executeOCLQueryWithoutContext(
					models.getRepositories().toArray(new EObject[ models.getRepositories().size()]),
					"self.oclIsKindOf(identifier::Identifier) and self.oclAsType(identifier::Identifier).id='" + identifier + "'");
		}
		assert queryResults != null && queryResults.size() < 2; // at most one service can be found
		return queryResults.isEmpty() ? null : queryResults.get(0);
	}

	private edu.kit.ipd.descartes.solving.probes.Probe mapDQLProbeToDMLProbe(
				tools.descartes.dql.models.mapping.domain.Probe dqlProbe) {
		if (DMLHelper.METRIC_RESPONSE_TIME_AVG.equalsIgnoreCase(dqlProbe.getMetricName())) {
			return ProbesFactory.eINSTANCE.createMeanResponseTimeProbe();
		} else if (DMLHelper.METRIC_UTILIZATION.equalsIgnoreCase(dqlProbe.getMetricName())) {
			return ProbesFactory.eINSTANCE.createAvgUtilizationProbe();
		} else if (DMLHelper.METRIC_THROUGHPUT.equalsIgnoreCase(dqlProbe.getMetricName())) {
			return ProbesFactory.eINSTANCE.createAvgThroughputProbe();
		}
		// TODO add support for resptime distribution probes
		// ProbesFactory.eINSTANCE.createResponseTimeDistributionProbe();
		else {
			log.warn("Could not map DQL probe " + dqlProbe.getMetricName() + " to corresponding DQL Probe.");
		}
		return null;
	}

	@Override
	public void setConstraint(final String constraint) {
		for (int i = 0; i < SUPPORTED_CONSTRAINTS.length; i++) {
			if (SUPPORTED_CONSTRAINTS[i].equalsIgnoreCase(constraint))
				this.constraintIndex = i;
		}
	}

	@Override
	public void setDoFConstraint(final String doFConstraint) {
		if (doFConstraintNames.contains(doFConstraint)) {
			this.doFConstraint = doFConstraint;
		} else {
			this.doFConstraint = null;
		}
	}

	@Override
	public void setDoFExplorationStrategy(final String strategy,
			final Properties properties) {
		// Unsupported
	}

	@Override
	public void setObservationPeriod(final Date start, final Date end,
			final long sampleSize) {
		// Unsupported
	}

	@Override
	public void setObservationPeriod(final long amount, final String unit,
			final long sampleSize) {
		// Unsupported
	}

	@Override
	public void setObservationReference(final String type,
			final String identifier, final long sampleSize) {
		// Intentionally left empty
	}

	private ProbeRepository startDMLSolver(final ProbeRepository probes, DMLModels models, Properties properties, SimulationConfig config) {		
		DMLSolver solver = new DMLSolver(models, config);
		return solver.solve();
	}
	
	private void saveProbesModel(EObject model, URI targetFolder) {
		if (null != targetFolder) {
			File file = new File(targetFolder.toFileString());
			if ((file.isDirectory()) ||  
				(!file.exists() && file.mkdir())) {
				String resourcefilename = "inputprobes" + System.currentTimeMillis() + ".probes";
				org.eclipse.emf.ecore.resource.Resource resource = 
						new ResourceSetImpl().createResource(targetFolder.appendSegment(resourcefilename));
				resource.getContents().add(model);
				try {
					resource.save(Collections.EMPTY_MAP);
				} catch (IOException e) {
					log.error(e, e.getCause());
				}
			}
		}
	}
	
	@Override
	public boolean supportsConstraintRequests() {
		return true;
	}

	@Override
	public boolean supportsDoFRequests() {
		return true;
	}

	@Override
	public boolean supportsMetricRequests() {
		return true;
	}

	@Override
	public boolean supportsObservationPeriods() {
		return false;
	}

	@Override
	public void reset() {
		// Nothing to do here
	}

	@Override
	public void addQueryProgressListener(QueryProgressListener arg0) {
		// TODO Auto-generated method stub
		
	}

}
